
#include "simpleTokenizer.h"
// The interface information of simpleScanner.cpp is in "simpleScanner.h", such as
// 		the prototypes of the static member functions of the TokenAnalyzer class.
// The caller of the static member functions of the TokenAnalyzer class
//		needs to include "simpleScanner.h"

using namespace std;

#include <iostream>  	// for basic iostream operations
#include <vector>		// the vector class is used in this implementation

//***************************************************************************
//*  getTokensFromOneCharArray
//*
//***************************************************************************
//*  This method analyzes one C-style character array or a string object and
//*  collectd the token strings into a vector of strings
//***************************************************************************
// 1. One line of code is stored in a C-style character array: ptrOriginalStatement,
//    with a terminating '\0' NULL character to end the string stored there.
// 2. We then store all the tokens, each as a string object of the string class,
//    into tokenArray, which is a vector of strings.

void TokenAnalyzer::getTokensFromOneCharArray
						(
							const char* ptrOriginalStatement,
							vector<string> & tokenVector
						)
{
 string statement;
 while ( (*ptrOriginalStatement) != NULL)
 {
	if ( isspace(*ptrOriginalStatement) )
		//Convert every white space characters into a space ' '
		//	and append it to the end of statement;
		statement = statement + ' ';
	else
		//Append a character as it is to the end of statement;
		statement = statement + (*ptrOriginalStatement);
	++ptrOriginalStatement;
 }


 string tokenBuffer;
 tokenBuffer.reserve(200);
 //reserve an initial storage of 200 characters

 const char DOUBLE_QUOTE = '\"';
 size_t iBuffIndex = 0; // index into the statement string

 //A while loop that normalizes the statement string
 //into an equaivalent string with one space between two adjacent tokens.
 while (iBuffIndex< statement.size())
 //While it is not the end of thecharacter string stored in the statement array
 {
	char c, nextChar;
	c = statement[iBuffIndex];
	nextChar = statement[iBuffIndex+1];

	if (c == DOUBLE_QUOTE)
	 //if a double quote is encountered, search for the matching double quote
	 //and skip the entire string literal like "Hello" enclosed in
	 //the pair of double quotes.
	{	tokenBuffer += DOUBLE_QUOTE;
		//Copy the starting double quote into token buffer

		iBuffIndex++;
		c = statement[iBuffIndex];
		while (c != NULL && c != DOUBLE_QUOTE)
		//Search for a matching double quote or a terminating NULL character.
		//Append all achracters in between into tokenBuffer
		 {	tokenBuffer += c;

			iBuffIndex++;
			c = statement[iBuffIndex];
		 }
		tokenBuffer += DOUBLE_QUOTE;
		//Copy the endinging double quote into token buffer
	}
	else //The character examined is not a double quote.
    // In the following id statement,
	// we try to detect the end (last character) of a token in the statement array.
	// We insert a space appropriately into the components string to make sure
	// in the components string there is at least one space between two tokens.
		if	(
				( (  isalnum(c) || c=='.' ) &&
					 ! (isalnum(nextChar) || nextChar=='.' )
			    )
				// If the current character c is alphnumerical character or a dot
				// (i.e. a letter ot a digit), but the next character is not.
				||
				( (c == '=') && ( nextChar != '=') )
				// equal sign, but not followed by another equal sign
				||
				c==',' ||c==':' || c==';'
				// comma, colon, semicolon
				||
				c=='+' || c=='-' || c=='*'
				// numerical operators: +,-, *
				||
				( (c == '/') && ( nextChar != '/') )
				// numerical operator: / but not followed by /
				||
				( (c == '>') && ( nextChar != '=') )
				// relational operator: > but not followed by =
				||
				( (c == '<') && ( nextChar != '=') )
				// relational operator: < but not followed by =
				||
				( (c == '&') && ( nextChar != '&') )
				// end of logical operator: &&
				||
				( (c == '|') && ( nextChar != '|') )
				// end of logical operator: ||
				||
				( (c == '!') )
				// end of logical operator: !
				||
				(	c== '(' || c==')'	)
				// parenthesis
				||
				(	c== '{' || c=='}'	)
				// curly braces
			)
			//We detect c as the end of a token.
			{
				// Copy the character c from statement array i
				// and append it into the end of of the token components string.
				tokenBuffer += c;

				// If the next character is not a space,
				// append a space to the end of components to make sure
				// in components there is at least one space between two tokens.
				if ( nextChar != ' ')
					tokenBuffer += ' ';
			}
		else
			// c is not the end of a token.
		{	if (c != ' ')
					// c is not a space, but a part of a token while not the end of the token.
					// So just copy the character c from statement array i
					// and append it into the end of of the token components string.
					tokenBuffer += c;
			else
			{	if (c == ' ' && nextChar != ' ')
					// c is a space, but the next character is not a space.
					// append a space to the end of components to make sure
					// in components there is at least one space between two tokens.
					tokenBuffer += ' ';
			}
		}

	if (c != NULL)
	//If the character we just finish examining is not the terminating NULL,
	//prepare to examine the next character on the next iteration.
			iBuffIndex++;
	//else we have reached a NULL character signaling the end of the statement.

	//Prepare to check the next character in the statement array.
 }
 //end of while loop that normalizes the statement in the character array
 //into an equaivalent string with one space between two adjacent tokens.


 size_t startIndex; // The index of the beginning of a token
 size_t endIndex;	// The index of the end of a token

 bool keepChecking;
 //Detect the first NON-SPACE character in the string.
 //Set startIndex as the index this character or
 //set it as the length of the string if the entire string is composed of spaces.
 for (	startIndex=0, keepChecking = true;
		startIndex < tokenBuffer.size() && keepChecking ;
	 )
	 if  ( tokenBuffer[startIndex]!= ' ')
		keepChecking = false;
	 else
		startIndex++;

 //***********************************************************
 //Collecting the token strings into the token vector
 //***********************************************************
 tokenVector.clear();

 while ( startIndex < tokenBuffer.size() )
 {
	if (tokenBuffer[startIndex]== DOUBLE_QUOTE)
	//If the token is a string LITERAL like "YES" started with a double quote,
	//search for the matching double quote after the index startIndex.
	//Set endIndex as the index after that of the matching double quote.
	{
		for (	endIndex = startIndex+1, keepChecking = true;
				endIndex < tokenBuffer.size() && keepChecking ;
				endIndex++
			)
			if  ( tokenBuffer[endIndex]== DOUBLE_QUOTE)
				keepChecking = false;
	}
	else
	//If the token is not started with a double quote,
	//Detect the first SPACE character after the index startIndex.
	//Set endIndex as the index this space character or
	//set it as the length of the string
	//if none such space exists after the index startIndex.
	{
		for (	endIndex = startIndex+1, keepChecking = true;
				endIndex < tokenBuffer.size() && keepChecking ;
			)
			if  ( tokenBuffer[endIndex]== ' ')
				keepChecking = false;
			else
				endIndex++;
	}

	//tokenBuffer.substr(startIndex, endIndex - startIndex) is a token.
	//Append such a token in the token vector to tokenVector
	tokenVector.push_back(tokenBuffer.substr(startIndex, endIndex - startIndex));
	//cout << "Token found:" << tokenVector.back() << endl;

	//Detect the next non-space character with an index after current endIndex.
	//Set startIndex as the index this character or
	//set it as the length of the string if the reaminder of the string
	//is entirely composed of spaces.
	for (	startIndex=endIndex, keepChecking = true;
			startIndex < tokenBuffer.size() && keepChecking ;
		)
		if  ( tokenBuffer[startIndex]!= ' ')
			keepChecking = false;
		else
			startIndex++;

 }

}


//*************************************************************************
// Display the tokens (and their category) in a vetor of tokens (strings)
//*************************************************************************
void TokenAnalyzer::displayTokens(vector<string> tokensOfExpression)
{		size_t numTokens = tokensOfExpression.size();

		cout << "***************************************************" << endl
			 << "The tokens of the arithmetic expression are:" << endl;
		for (int i=0; i < numTokens; i++)
			cout << "[Token " << i << "]=" << tokensOfExpression[i] << endl;

		cout << endl << "Check the tokens one by one:" << endl;
		for (int i=0; i < numTokens; i++)
		{
			if ( isNUMERICAL_LITERAL( tokensOfExpression[i] ) )
				cout	<< "Numerical constant:\t [Token " << i << "]=" << tokensOfExpression[i] 
						<< "\t numerical value =" << atof(tokensOfExpression[i].c_str()) 
						<< endl;

			if ( isNUMERICAL_OP( tokensOfExpression[i] ) )
				cout	<< "Numerical operator:\t [Token " << i << "]=" << tokensOfExpression[i] 
						<< endl;

			if ( isLEFT_PARENTHESIS( tokensOfExpression[i] ) )
				cout	<< "Left parenthesis:\t [Token " << i << "]=" << tokensOfExpression[i] 
						<< endl;

			if ( isRIGHT_PARENTHESIS( tokensOfExpression[i] ) )
				cout	<< "Right parenthesis:\t [Token " << i << "]=" << tokensOfExpression[i] 
						<< endl;
		}


}


//******************************************************************
//* The following is a colllection of functions to
//* test the categories of tokens
//******************************************************************

//Check whether a token is a numerical constant
bool TokenAnalyzer::isNUMERICAL_LITERAL(const string & token)
{
	if (token == ".")
		return false;

	bool decimalPointSeenBefore = false;
	//no decimal points seen yet

	for (	size_t i=0; //index into the string
			i < token.size();
			i++	)
		{
			char c;
			c = token[i];
			if ( c != '.' && !isdigit(c) )
				// not a digit, nor a demical point ==> Not Numerical
				return false;
			if ( c == '.' && decimalPointSeenBefore)
				// demical point appears more than once ==> Not right
				return false; //Not Numerical
			if ( c == '.' && !decimalPointSeenBefore)
				// Keep a record when decimal point is seen for the first time
				decimalPointSeenBefore = true;

		}

	return true;
}


bool TokenAnalyzer::isNUMERICAL_OP(const string & token)
// Test whether it is one of the following
// numerical operators	 (such as +, -, *, /)
{
	if	(	token == "+"	||
			token == "-"	||
			token == "*"	||
			token == "/"	
	    )
		return true;
	else
		return false;
}


bool TokenAnalyzer::isLEFT_PARENTHESIS(const string & token)
// Test whether it is a left parenthesis (
{
	if (token == "(")
		return true;
	else
		return false;
}

bool TokenAnalyzer::isRIGHT_PARENTHESIS(const string & token)
// Test whether it is a right parenthesis )
{
	if (token == ")")
		return true;
	else
		return false;
}
